
; una "lista associativa" e' una struttura dati che rappresenta una
; associazione tra nomi e valori
;
; si implementa mediante una lista di coppie <nome, valore>, chiamamo
; ognuna di queste coppie "binding" (legame)
(define-struct binding (name value))

; lookup (ricerca) di un nome all'interno di una lista associativa. Se
; il nome richiesto e' legato all'interno della lista, restituisce il
; valore associato, altrimenti ritorna un messaggio di errore
(define (lookup name l)
  (cond
    [(empty? l)
     (error 'lookup
            (string-append
             (symbol->string name)
             " "
             "not found"))]
    [(symbol=? (binding-name (first l)) name)
     (binding-value (first l))]
    [else (lookup name (rest l))]))

; lista associativa di esempio:
; associa 1 al simbolo 'a e 2 al simbolo
(define lista1 (list (make-binding 'a 1) (make-binding 'b 2)))

; esempio di lookup:
;(lookup 'a lista1)

; un albero di espressioni e':
; - un numero
; - la somma di due alberi di espressioni (struttura "piu")
; - il prodotto di due alberi di espressioni  (struttura "per")
(define-struct piu (arg1 arg2))
(define-struct per (arg1 arg2))

;esempio
;rappresentazione dell'espressione scheme (+ (* 3 4) 5)
;      +
;    /   \
;   *     5
;  / \
; 3   4
(define tree1 (make-piu (make-per 3 4) 5))

; funzione di valutazione: dato un albero di espressioni in input,
; calcola il valore corrispondente
(define (eval1 tree)
  (cond
    [(number? tree) tree]
    [(piu? tree)
     (+ (eval1 (piu-arg1 tree))
        (eval1 (piu-arg2 tree)))]
    [(per? tree)
     (* (eval1 (per-arg1 tree))
        (eval1 (per-arg2 tree)))]))

;esempio di eval1:
;(eval1 tree1)

;definizione alternativa di alberi di espressioni, un albero di
;espressioni e':
; - un numero
; - un operatore binario

;definizione dell'operatore binario, name e' un simbolo corrispondente
;all'operatore utilizzato: 'piu per la somma, 'per per la
;moltiplicazione
(define-struct op (name arg1 arg2))

;esempio
;rappresentazione dell'espressione scheme (+ (* 3 4) 5)
;      +
;    /   \
;   *     5
;  / \
; 3   4
(define tree2 (make-op 'piu (make-op 'per 3 4) 5))

;valutazione per la definizione alternativa
(define (eval2 tree)
  (cond
    [(number? tree) tree]
    [(symbol=? 'piu (op-name tree))
     (+ (eval2 (op-arg1 tree))
        (eval2 (op-arg2 tree)))]
    [(symbol=? 'per (op-name tree))
     (* (eval2 (op-arg1 tree))
        (eval2 (op-arg2 tree)))]))
;esempio di eval2:
;(eval2 tree2)

;aumentiamo la definizione di alberi di espressioni in modo che le
;foglie non siano solo numeri ma anche simboli. La foglia 'x (simbolo x)
;rappresenta l'occorrenza di una variabile nell'espressione che vogliamo
;codificare. Per valutare tali alberi abbiamo bisogno di un ambiente,
;ovvero di una lista associativa che mi associ ad ogni variabile che
;occorre nell'albero una espressione

;esempio, rappresentazione dell'epressione scheme (+ (* 1 2) x)
;      +
;    /   \
;   *     'x
;  / \
; 1   2
(define tree3 (make-op 'piu (make-op 'per 1 2) 'x))

;esempio di ambiente che associa 10 a 'x
(define env3 (list (make-binding 'x 10)))
;esempio di ambiente che associa (+ 6 4) a 'x
(define env3bis (list (make-binding 'x 10)))

;funzione di valutazione per alberi di espressioni contenenti variabili
(define (eval3 env tree)
  (cond
    [(number? tree) tree]
    [(op? tree)
     (cond
       [(symbol=? 'piu (op-name tree))
        (+ (eval3 env (op-arg1 tree))
           (eval3 env (op-arg2 tree)))]
       [(symbol=? 'per (op-name tree))
        (* (eval3 env (op-arg1 tree))
           (eval3 env (op-arg2 tree)))])]
     [(symbol? tree) 
      (eval3 env (lookup tree env))]))

;esempi di eval3:
;(eval3 env3 tree3)
;(eval3 env3bis tree3)

;aumentiamo la definizione di alberi di espressioni in modo che possano
;contenere anche applicazioni di funzioni unarie
;
;una funzione e' una tripla <nome, argomento, corpo>
;dove nome e' un simbolo rappresentante il nome della funzione,
;argomento un simbolo rappresentante il nome del parametro della
;funzione e corpo un albero di espressioni che puo' contenere il
;parametro della funzione
(define-struct fun (name arg body))

;esempio di funzione, rappresentazione della definizione scheme
;(define (double x) (+ x x))
(define fun1 (make-fun 'double 'x (make-op 'piu 'x 'x)))

;esempio, rappresentazione dell'espressione scheme (+ (double 2) 1)
;L'applicazione di funzione e' rappresentata come un operatore 'appl
;avente come primo argomento il nome della funzione e come secondo
;l'argomento della funzione
(define expr4 (make-op 'piu (make-op 'appl 'double 2) 1))

;esempio, ambiente che lega il nome 'double alla funzione fun1 definita
;sopra
(define env4 (list (make-binding 'double fun1)))

;funzione di valutazione per alberi di espressioni contenenti numeri,
;variabili e applicazioni di funzioni
(define (eval4 env tree)
  (cond
    [(number? tree) tree]
    [(op? tree)
     (cond
       [(symbol=? 'piu (op-name tree))
        (+ (eval4 env (op-arg1 tree))
           (eval4 env (op-arg2 tree)))]
       [(symbol=? 'per (op-name tree))
        (* (eval4 env (op-arg1 tree))
           (eval4 env (op-arg2 tree)))]
       [(symbol=? 'appl (op-name tree))
        (eval4
        (cons (make-binding
               (fun-arg (lookup (op-arg1 tree) env))
               (eval4 env (op-arg2 tree)))
              env)
        (fun-body (lookup (op-arg1 tree) env)))])]
     [(symbol? tree) 
      (eval4 env (lookup tree env))]))

;esempi di eval4:
;(eval4 env4 expr4)

